from typing import List


class AppRequestInfo:
    def __init__(self, name: str, instances_count: int, app_dependencies: List[str] = None):
        self.name = name
        self.instances_count = instances_count
        self.app_dependencies = app_dependencies if app_dependencies is not None else []
